import dataclasses
from enum import Enum, EnumMeta
from typing import List, Any

from railib import api

from workflow.constants import ACCOUNT_PARAM, CONTAINER_PARAM, DATA_PATH_PARAM, AZURE_SAS, CONTAINER, CONTAINER_TYPE, \
    CONTAINER_NAME, USER_PARAM, PASSWORD_PARAM, SNOWFLAKE_ROLE, SNOWFLAKE_WAREHOUSE, DATABASE_PARAM, SCHEMA_PARAM, \
    FAIL_ON_MULTIPLE_WRITE_TXN_IN_FLIGHT, RAI_SDK_HTTP_RETRIES, RAI_PROFILE, RAI_PROFILE_PATH, \
    SEMANTIC_SEARCH_BASE_URL, RAI_CLOUD_ACCOUNT


class MetaEnum(EnumMeta):
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


class FileFormat(str, BaseEnum):
    CSV = '.csv'
    JSON = '.json'
    JSONL = '.jsonl'
    CSV_GZ = '.csv.gz'
    JSON_GZ = '.json.gz'
    JSONL_GZ = '.jsonl.gz'

    @staticmethod
    def is_supported(file_name):
        """
        Check if file has supported format
        :param file_name: File name
        :return: True if a file name ends with supported file format
        """
        for f in FileFormat:
            if file_name.endswith(f):
                return True
        return False


class FileType(str, BaseEnum):
    CSV = 'CSV'
    JSON = 'JSON'
    JSONL = 'JSONL'


class ContainerType(str, BaseEnum):
    LOCAL = 'local'
    AZURE = 'azure'
    SNOWFLAKE = 'snowflake'

    def __str__(self):
        return self.value

    @staticmethod
    def from_source(src):
        try:
            return ContainerType[src["container_type"]]
        except KeyError as ex:
            raise ValueError(f"Container type is not supported: {ex}")


@dataclasses.dataclass
class FileMetadata:
    path: str
    size: int = None
    as_of_date: str = ""


@dataclasses.dataclass
class Container:
    name: str
    type: ContainerType
    params: dict[str, Any]


@dataclasses.dataclass
class AzureConfig:
    account: str
    container: str
    data_path: str
    sas: str


@dataclasses.dataclass
class LocalConfig:
    data_path: str


@dataclasses.dataclass
class SnowflakeConfig:
    account: str
    user: str
    password: str
    role: str
    warehouse: str
    database: str
    schema: str


class ConfigExtractor:

    @staticmethod
    def azure_from_env_vars(env_vars: dict[str, Any]):
        return AzureConfig(
            account=env_vars.get(ACCOUNT_PARAM, ""),
            container=env_vars.get(CONTAINER_PARAM, ""),
            data_path=env_vars.get(DATA_PATH_PARAM, ""),
            sas=env_vars.get(AZURE_SAS, "")
        )

    @staticmethod
    def snowflake_from_env_vars(env_vars: dict[str, Any]):
        return SnowflakeConfig(
            account=env_vars.get(ACCOUNT_PARAM, ""),
            user=env_vars.get(USER_PARAM, ""),
            password=env_vars.get(PASSWORD_PARAM, ""),
            role=env_vars.get(SNOWFLAKE_ROLE, ""),
            warehouse=env_vars.get(SNOWFLAKE_WAREHOUSE, ""),
            database=env_vars.get(DATABASE_PARAM, ""),
            schema=env_vars.get(SCHEMA_PARAM, "")
        )

    @staticmethod
    def local_from_env_vars(env_vars: dict[str, Any]):
        return LocalConfig(data_path=env_vars.get(DATA_PATH_PARAM, ""))


@dataclasses.dataclass
class Source:
    container: Container
    relation: str
    relative_path: str
    input_format: str
    extensions: List[str]
    is_chunk_partitioned: bool
    is_date_partitioned: bool
    loads_number_of_days: int
    offset_by_number_of_days: int
    snapshot_validity_days: int
    paths: List[str] = dataclasses.field(default_factory=list)

    def to_paths_csv(self) -> str:
        return "\n".join([f"{self.relation},{self.container.name},{p}" for p in self.paths])

    def to_chunk_partitioned_paths_csv(self) -> str:
        return "\n".join([f"{self.relation},{path},{self.is_chunk_partitioned}" for path in self.paths])

    def to_formats_csv(self) -> str:
        return f"{self.relation},{self.input_format.upper()}"

    def to_container_type_csv(self) -> str:
        return f"{self.relation},{self.container.type.name}"

    def is_size_supported(self) -> bool:
        return self.container.type == ContainerType.LOCAL or self.container.type == ContainerType.AZURE


@dataclasses.dataclass
class RaiConfig:
    ctx: api.Context
    engine: str
    database: str


@dataclasses.dataclass
class EnvConfig:
    containers: dict[str, Container]
    fail_on_multiple_write_txn_in_flight: bool = False
    rai_sdk_http_retries: int = 3
    rai_profile: str = "default"
    rai_profile_path: str = "~/.rai/config"
    semantic_search_base_url: str = ""
    rai_cloud_account: str = ""

    __EXTRACTORS = {
        ContainerType.AZURE: lambda env_vars: ConfigExtractor.azure_from_env_vars(env_vars),
        ContainerType.LOCAL: lambda env_vars: ConfigExtractor.local_from_env_vars(env_vars),
        ContainerType.SNOWFLAKE: lambda env_vars: ConfigExtractor.snowflake_from_env_vars(env_vars)
    }

    def get_container(self, name: str) -> Container:
        try:
            return self.containers[name]
        except KeyError:
            raise ValueError(f"Container `{name}` is missed in Environment Config.")

    @staticmethod
    def get_config(container: Container):
        return EnvConfig.__EXTRACTORS[container.type](container.params)

    @staticmethod
    def from_env_vars(env_vars: dict[str, Any]):
        containers = {}
        for container in env_vars.get(CONTAINER, []):
            name = container[CONTAINER_NAME]
            containers[name] = Container(name=container[CONTAINER_NAME],
                                         type=ContainerType[container[CONTAINER_TYPE].upper()],
                                         params=container)
        return EnvConfig(containers, env_vars.get(FAIL_ON_MULTIPLE_WRITE_TXN_IN_FLIGHT, False),
                         env_vars.get(RAI_SDK_HTTP_RETRIES, 3), env_vars.get(RAI_PROFILE, "default"),
                         env_vars.get(RAI_PROFILE_PATH, "~/.rai/config"), env_vars.get(SEMANTIC_SEARCH_BASE_URL, ""),
                         env_vars.get(RAI_CLOUD_ACCOUNT, ""))


@dataclasses.dataclass
class Export:
    meta_key: List[str]
    relation: str
    relative_path: str
    file_type: FileType
    snapshot_binding: str
    container: Container
    offset_by_number_of_days: int = 0
    is_partitioned: bool = False


@dataclasses.dataclass
class BatchConfig:
    name: str
    content: str



import asyncio
import logging
import os
import time
import yaml
import json
from schema import Schema
from types import MappingProxyType
from workflow.schema import Validator

from datetime import datetime, timedelta
from typing import List, Dict

from workflow import constants
from workflow.common import LocalConfig
from workflow.exception import RetryException


def range_days(start: datetime, end: datetime) -> List[datetime]:
    delta = end - start
    return [start + timedelta(days=i) for i in range(delta.days + 1)]


def parse_date(date_str: str) -> datetime:
    return datetime.strptime(date_str, constants.DATE_FORMAT)


def get_date_path(end_date: str, date_format: str, day_offset: int):
    date = datetime.strptime(end_date, date_format) - timedelta(days=day_offset)
    return f"{constants.DATE_PREFIX}{date.strftime(date_format)}"


def read(fname: str) -> str:
    with open(fname) as fp:
        return fp.read()


def read_config(fname: str, schemas: dict[str, Schema] = MappingProxyType({})) -> str:
    _, file_extension = os.path.splitext(fname)
    if file_extension == ".json":
        json_string = read(fname)
        config_data = json.loads(json_string)
    elif file_extension == ".yaml" or file_extension == ".yml":
        with open(fname, 'r') as yaml_file:
            config_data = yaml.safe_load(yaml_file)
        json_string = json.dumps(config_data, indent=2)
    else:
        raise Exception(f"Unsupported batch config file extension: {file_extension}")
    Validator(schemas).validate(config_data)
    return json_string


def sansext(fname: str) -> str:
    return os.path.splitext(os.path.basename(fname))[0]


def save_csv_output(outputs: Dict, config: LocalConfig) -> None:
    """
    Save the content of dictionary as CSV files
    :param outputs:     dictionary with outputs
    :param config:      local config
    :return:
    """
    for output in outputs.keys():
        # for the time being, this handles the specialized relations of meta-exports
        normalized_file_name = output.replace("/:", "_")
        with open(f"{config.data_path}/{normalized_file_name}.csv", "w") as file:
            file.write(outputs[output])


def build_relation_path(relation: str, *keys: str) -> str:
    """
    Build relation from base and paths.
    Ex. base:`batch:config`, paths: ['daily', 'fake'] => `batch:config:daily:fake`
    :param relation:    base relation
    :param keys:        path parts
    :return: new relation
    """
    relation_path = ":" + ":".join(keys) if len(keys) > 0 else ""
    return f"{relation}{relation_path}"


def build_models(filenames: List[str], files_root: str) -> dict:
    """
    Build RAI models from list of files.
    :param filenames:   file names
    :param files_root:  files root path
    :return:
    """
    models = {}
    for file in filenames:
        filename = f"{files_root}/{file}"
        with open(filename) as fp:
            models[file] = fp.read()
    return models


def format_duration(seconds: float) -> str:
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    hours_int = int(hours)
    minutes_int = int(minutes)
    seconds_int = int(seconds)
    if hours_int != 0:
        return f"[{hours_int:d}h {minutes_int:d}m {seconds_int:d}s]"
    elif minutes_int != 0:
        return f"[{minutes_int:d}m {seconds_int:d}s]"
    else:
        return f"[{seconds_int:d}s]"


def get_common_model_relative_path(file) -> str:
    """
    Get relative path to common model from folder of given file.
    :param file:    file
    :return:
    """
    return os.path.dirname(os.path.realpath(file)) + constants.COMMON_MODEL_RELATIVE_PATH


def extract_date_range(logger: logging.Logger, start_date_str, end_date_str, number_of_days, offset_of_days) ->\
        List[str]:
    end_date = parse_date(end_date_str)
    offset = offset_of_days if offset_of_days is not None else 0
    end_date = end_date - timedelta(offset)
    start_date_adjusted = end_date - timedelta(number_of_days - 1) if number_of_days is not None else None
    if start_date_str:
        start_date = parse_date(start_date_str)
        start_date = start_date_adjusted if start_date_adjusted is not None and start_date < start_date_adjusted\
            else start_date
    else:
        start_date = start_date_adjusted
    start_date = start_date if start_date is not None else end_date
    logger.info(f"Building range from '{start_date}' to '{end_date}'")
    return [date.strftime(constants.DATE_FORMAT) for date in range_days(start_date, end_date)]


def to_rai_date_format(date_format: str) -> str:
    fmt_part_map = {
        "%Y": "YYYY",
        "%m": "mm",
        "%d": "dd"
    }
    rai_date_format = date_format
    for py_fmt, rai_fmt in fmt_part_map.items():
        rai_date_format = rai_date_format.replace(py_fmt, rai_fmt)
    return rai_date_format


async def call_with_overhead_async(
        f,
        logger: logging.Logger,
        overhead_rate: float,
        start_time: int = time.time(),
        timeout: int = None,
        max_tries: int = None,
        first_delay: float = 0.5,
        max_delay: int = 120,  # 2 minutes
) -> None:
    tries = 0
    max_time = time.time() + timeout if timeout else None

    while True:
        logger.debug(f"Calling function. The number of try: {tries + 1}")
        if f():
            break

        if max_tries is not None and tries >= max_tries:
            raise RetryException(f'max tries {max_tries} exhausted')

        if max_time is not None and time.time() >= max_time:
            raise RetryException(f'timed out after {timeout} seconds')

        tries += 1
        duration = min((time.time() - start_time) * overhead_rate, max_delay)
        if tries == 1:
            logger.debug(f"Sleep duration for the first try: {first_delay}s")
            await asyncio.sleep(first_delay)
        else:
            logger.debug(f"Sleep duration for a try: {duration}s")
            await asyncio.sleep(duration)


def call_with_overhead(
        f,
        logger: logging.Logger,
        overhead_rate: float,
        start_time: int = time.time(),
        timeout: int = None,
        max_tries: int = None,
        first_delay: float = 0.5,
        max_delay: int = 120,  # 2 minutes
):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(
        call_with_overhead_async(f, logger, overhead_rate, start_time, timeout, max_tries, first_delay, max_delay))


def get_or_create_eventloop():
    """Get the current event loop or create a new one."""
    try:
        return asyncio.get_event_loop()
    except RuntimeError as ex:
        if "There is no current event loop in thread" in str(ex):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return asyncio.get_event_loop()
        raise ex



import logging

import requests
from workflow.exception import RestClientException


class RestClient:
    def __init__(self, logger: logging.Logger, base_url: str):
        self.logger = logger
        self.base_url = base_url

    def _make_request(self, method, endpoint, data=None, headers=None, files=None):
        url = f"{self.base_url}/{endpoint}"
        try:
            response = requests.request(method, url, data=data, headers=headers, files=files)
            response.raise_for_status()  # Raise an HTTPError for bad responses (4xx or 5xx)
            return response
        except requests.exceptions.HTTPError as http_err:
            self.logger.error(f"HTTP error occurred: {http_err}")
            raise RestClientException(method, url, http_err)
        except Exception as err:
            self.logger.error(f"An error occurred: {err}")
            raise RestClientException(method, url, err)

    def _make_json_request(self, method, endpoint, data=None, headers=None, files=None):
        return self._make_request(method, endpoint, data=data, headers=headers, files=files).json()

    def _make_file_request(self, method, endpoint, data=None, headers=None, files=None):
        return self._make_request(method, endpoint, data=data, headers=headers, files=files).content

    def get(self, endpoint, headers=None):
        return self._make_json_request("GET", endpoint, headers=headers)

    def get_file_content(self, endpoint, headers=None):
        return self._make_file_request("GET", endpoint, headers=headers)

    def post(self, endpoint, data=None, headers=None, files=None):
        return self._make_json_request("POST", endpoint, data=data, headers=headers, files=files)

    def delete(self, endpoint, headers=None):
        return self._make_request("DELETE", endpoint, headers=headers)

    def put(self, endpoint, data=None, headers=None, files=None):
        return self._make_json_request("PUT", endpoint, data=data, headers=headers, files=files)

